unit FBErrorsFireDAC;

interface

uses
  // general uses
  System.SysUtils, Data.DB,
  // FireDAC uses
  uADCompClient,

  FBMetaData;

type
  TFBErrors = class
  private
    FE: Exception;
    FMeta: TFBMetaData;
    FDataBase: TCustomConnection;
  protected
    function GetGeneralMsg(GeneralMessage, MessageError: string; ErrorCode: Integer): string; virtual;
    function GetDuplicatePrimUnqMsg(MsgError: string; ErrorCode: Integer): string; virtual;
    function GetIOErrorMsg(MsgError: string; ErrorCode: Integer): string; virtual;
    function GetUnavailableMsg(MsgError: string; ErrorCode: Integer): string; virtual;
    function GetSQLErrorMsg(MsgError: string; ErrorCode: Integer): string; virtual;
    function GetForeingKeyMsg(MsgError: string; ErrorCode: Integer): string; virtual;
    function GetConstraintCheckMsg(MsgError: string; ErrorCode: Integer): string; virtual;

    function GetErrorCode: Integer; virtual; abstract;
    procedure CreateMeta; virtual; abstract;

    function GetMessageError: string; virtual;
  public
    constructor Create(E: Exception; DataBase: TCustomConnection); virtual;
  end;

  TFBErrorsFireDAC = class(TFBErrors)
  protected
    function GetErrorCode: Integer; override;
    procedure CreateMeta; override;

    function GetMessageError: string; override;
  public
    constructor Create(E: Exception; DataBase: TADConnection); reintroduce; virtual;

    class function GetFBMessageError(E: Exception; DataBase: TADConnection): string;
  end;

resourcestring
  errNotCoded = 'Error no codificado';
  errDeadLock = 'Punto muerto (deadlock)';

  // connect error
  errConnectingDB =
    '  Se ha producido un error conectando con la base de datos.'#13#13 +
    '  Base de datos: %s'#13 +
    '  Cdigo del error: %d';
  errUnavailable =
    '  No ha sido posible conectar con la base de datos.'#13#13 +
    '  Una posible causa es que no est en marcha el servicio de Firebird.';

  // SQL Error
  errSQLError =
    '  Hay un error en la sentencia SQL.'#13#13 +
    '  Cdigo del error: %d'#13#13 +
    '  Mensaje del error:'#13 +
    '  %s';

  // Post Errors
  // General error
  errGeneralText =
    '  Se ha producido un error grabando la informacin. '#13#13 +
    '  Causa: %s '#13 +
    '  Informacin del error: %s '#13#13 +
    '  Cdigo del error: %d';

  errDuplicatePriUniq =
    '  Se ha producido un error en la grabacin del registro.'#13#13 +
    '  Causa: %s '#13 +
    '  Informacin extra: '#13 +
    '   - Tabla: %s '#13 +
    '   - ndice: %s '#13 +
    '   - Campos afectados: %s '#13#13 +
    'Modifique el/los valor/es e intntelo de nuevo. Si el problema persiste, contacte con el distribuidor.';

  errForeingKey =
    '  Se ha producido un error en la grabacin del registro.'#13#13 +
    '  Causa: %s '#13 +
    '  Informacin extra: '#13 +
    '   - Tabla: %s '#13 +
    '   - ndice: %s '#13 +
    '   - Campos afectados: %s '#13#13 +
    'Modifique el/los valor/es e intntelo de nuevo. Si el problema persiste, contacte con el distribuidor.';

  errConstraintCheck =
    '  Se ha producido un error en la grabacin del registro.'#13#13 +
    '  Causa: Restriccin con comprobacin '#13 +
    '  Informacin extra: '#13 +
    '   - Tabla: %s '#13 +
    '   - Constraint: %s '#13 +
    '   - Restriccin: %s '#13#13 +
    'Modifique el/los valor/es e intntelo de nuevo. Si el problema persiste, contacte con el distribuidor.';



  errValidationField =
    '  Se ha producido un error en la grabacin del registro. '#13#13 +
    '  Causa: %s '#13 +
    '  Generado por: %s '#13#13 +
    '  Informacin extra: %s ';

  errValidation =
    #13'   - Campo: %s '#13 +
       '   - Valor no vlido: %s '#13#13 +
       'Modifique el valor e intntelo de nuevo. Si el problema persiste, contacte con el distribuidor.';

implementation

uses
  // general uses
  System.Classes,

  // FireDAC uses
  uADPhysIBWrapper, uADPhysIBCli;

{ TFBErrors }

constructor TFBErrors.Create(E: Exception; DataBase: TCustomConnection);
begin
  FE := E;
  FDataBase := DataBase;
  CreateMeta;
end;

function TFBErrors.GetConstraintCheckMsg(MsgError: string;
  ErrorCode: Integer): string;
var
  Constraint,
  Table,
  Trigger,
  Condition: string;
  A, B: Integer;
begin
  // getting constraint name
  A := Pos('constraint', MsgError) + 10;
  B := Pos('on view', MsgError);
  Constraint := Trim(Copy(MsgError, A, B-A));

  // getting table name
  A := Pos('on view', MsgError) + 17;
  B := Pos('At trigger', MsgError);
  Table := Trim(Copy(MsgError, A, B-A));

  // getting table name
  A := Pos('''', MsgError, B+1) + 1;
  B := Length(MsgError);
  Trigger := Trim(Copy(MsgError, A, B-A));

  Condition := FMeta.GetConstraintCheck(Constraint, Trigger);

  Result := Format(errConstraintCheck, [Table,
                                        Constraint,
                                        Condition
                                        ]);
end;

function TFBErrors.GetDuplicatePrimUnqMsg(MsgError: string;
  ErrorCode: Integer): string;
var
  Constraint,
  Table,
  Tmp,
  Cols: string;
  A, B: Integer;
  L: TStringList;
  ConsType: TConstraintType;
begin
  // getting constraint name
  A := Pos('"', MsgError)+1;
  B := Pos('"', MsgError, A+1);
  Constraint := Trim(Copy(MsgError, A, B-A));

  // getting table name
  A := Pos('"', MsgError, B+1)+1;
  B := Pos('"', MsgError, A+1);
  Table := Trim(Copy(MsgError, A, B-A));

  L := TStringList.Create;
  try
    ConsType := FMeta.GetConstraintFields(Constraint, L);
    Cols := L.CommaText;
  finally
    FreeAndNil(L);
  end;

  case ConsType of
    ctUnique: Tmp := 'Clave nica duplicada.';
    ctPrimary: Tmp := 'ndice primario duplicado.';
    else Tmp := 'Clave duplicada.';
  end;

  Result := Format(errDuplicatePriUniq, [Tmp,
                                         Table,
                                         Constraint,
                                         Cols
                                        ]);
end;

function TFBErrors.GetForeingKeyMsg(MsgError: string;
  ErrorCode: Integer): string;
var
  Constraint,
  Table,
  Tmp,
  Cols: string;
  A, B: Integer;
  L: TStringList;
  ConsType: TConstraintType;
begin
  A := Pos('"', MsgError)+1;
  B := Pos('"', MsgError, A+1);
  Constraint := Trim(Copy(MsgError, A, B-A));

  A := Pos('"', MsgError, B+1)+1;
  B := Pos('"', MsgError, A+1);
  Table := Trim(Copy(MsgError, A, B-A));

  L := TStringList.Create;
  try
    ConsType := FMeta.GetConstraintFields(Constraint, L);
    Cols := L.CommaText;
  finally
    FreeAndNil(L);
  end;

  case ConsType of
    ctForeing: Tmp := 'Clave fornea inexistente.';
    else Tmp := 'Clave inexistente.';
  end;

  Result := Format(errDuplicatePriUniq, [Tmp,
                                         Table,
                                         Constraint,
                                         Cols
                                        ]);
end;

function TFBErrors.GetGeneralMsg(GeneralMessage, MessageError: string;
  ErrorCode: Integer): string;
begin
  Result := Format(errGeneralText, [GeneralMessage,
                                    MessageError,
                                    ErrorCode]);
end;

function TFBErrors.GetIOErrorMsg(MsgError: string; ErrorCode: Integer): string;
var
  DBase: string;
  L: TStringList;
begin
  L := TStringList.Create;
  try
    L.Delimiter := '"';
    L.StrictDelimiter := True;
    L.DelimitedText := MsgError;
    DBase := L[3];
  finally
    FreeAndNil(L);
  end;
  Result := Format(errConnectingDB, [DBase,
                                     ErrorCode]);
end;

function TFBErrors.GetMessageError: string;
begin
  case GetErrorCode of
    isc_io_error: // database unavailable
      Result := GetIOErrorMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);
    isc_unavailable: // Firebird service down
      Result := GetUnavailableMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);

    isc_lock_conflict: //
      Result := GetGeneralMsg(errDeadLock, EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);

    isc_dsql_error: // SQL sintax error
      Result := GetSQLErrorMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);

    isc_unique_key_violation: // duplicate primary or unique
      Result := GetDuplicatePrimUnqMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);
    isc_foreign_key: // clave fornea inexistente
      Result := GetForeingKeyMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);
    isc_check_constraint: // restriccin con check
      Result := GetConstraintCheckMsg(EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);
    isc_not_valid: begin end; // validation column
    else Result := GetGeneralMsg(errNotCoded, EIBNativeException(FE).Message, EIBNativeException(FE).ErrorCode);
  end;
end;

function TFBErrors.GetSQLErrorMsg(MsgError: string; ErrorCode: Integer): string;
begin
  Result := Format(errSQLError, [ErrorCode, MsgError]);
end;

function TFBErrors.GetUnavailableMsg(MsgError: string; ErrorCode: Integer): string;
begin
  Result := errUnavailable;
end;

{ TFBErrorsFireDAC }

constructor TFBErrorsFireDAC.Create(E: Exception; DataBase: TADConnection);
begin
  inherited Create(E, DataBase);
end;

procedure TFBErrorsFireDAC.CreateMeta;
begin
  FMeta := TFBMDFireDAC.Create(TADConnection(FDataBase));
end;

function TFBErrorsFireDAC.GetErrorCode: Integer;
begin
  Result := -1;
  if FE is EIBNativeException then
    Result := EIBNativeException(FE).ErrorCode;
end;

class function TFBErrorsFireDAC.GetFBMessageError(E: Exception;
  DataBase: TADConnection): string;
var
  FBError: TFBErrors;
begin
  FBError := TFBErrorsFireDAC.Create(E, DataBase);
  try
    Result := FBError.GetMessageError;
  finally
    FreeAndNil(FBError);
  end;
end;

function TFBErrorsFireDAC.GetMessageError: string;
begin
  Result := FE.Message;

  // slo controlamos el tipo de excepcin EIBNativeException
  if not (FE is EIBNativeException) then
  begin
    // miramos que no sea un error de base de datos EDatabaseError
    if not(FE is EDatabaseError) then
      Result := GetGeneralMsg(errNotCoded, FE.Message, -1)
    else
      Result := GetGeneralMsg(errNotCoded, EDatabaseError(FE).Message, -1);
    Exit;
  end;

  inherited;
end;

end.
